/*
* Copyright @ Members of the EMI Collaboration, 2010.
* See www.eu-emi.eu for details on the copyright holders.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
* 
* Authors : David Cameron, Alejandro Alvarez Ayllon, Adrien Devresse
* 
*/


#include <ctype.h>
#include <gfal_plugins_api.h>
#include <transfer/gfal_transfer.h>
#include <transfer/gfal_transfer_plugins.h>
#include <transfer/gfal_transfer_types.h>
#include "gfal_xrootd_plugin_interface.h"
#include "gfal_xrootd_plugin_utils.h"

#undef TRUE
#undef FALSE

#include <XrdCl/XrdClCopyProcess.hh>
#include <XrdVersion.hh>

#define XROOTD_DEFAULT_CHECKSUM "COPY_CHECKSUM_TYPE"



class CopyFeedback: public XrdCl::CopyProgressHandler {
public:
  CopyFeedback(gfal2_context_t context, gfalt_params_t p): context(context), params(p), startTime(0)
  {
    this->monitorCallback = gfalt_get_monitor_callback(this->params, NULL);
    monitorCallbackData = gfalt_transfer_status_create(&this->status);
  }

  virtual ~CopyFeedback()
  {
    gfalt_transfer_status_delete(this->monitorCallbackData);
  }


  void BeginJob(uint16_t jobNum, uint16_t jobTotal,
                const XrdCl::URL *source, const XrdCl::URL *destination)
  {
    this->startTime   = time(NULL);
    this->source      = source->GetURL();
    this->destination = destination->GetURL();

    plugin_trigger_event(this->params, xrootd_domain,
                         GFAL_EVENT_NONE, GFAL_EVENT_TRANSFER_ENTER,
                         "%s => %s",
                         this->source.c_str(), this->destination.c_str());
  }


  void EndJob(const XrdCl::XRootDStatus &status)
  {
    plugin_trigger_event(this->params, xrootd_domain,
                         GFAL_EVENT_NONE, GFAL_EVENT_TRANSFER_EXIT,
                         "%s", status.ToStr().c_str());
  }


  void JobProgress(uint64_t bytesProcessed, uint64_t bytesTotal)
  {
    if (this->monitorCallback) {
      time_t now = time(NULL);
      time_t elapsed = now - this->startTime;

      this->status.status           = 0;
      this->status.bytes_transfered = bytesProcessed;
      this->status.transfer_time    = elapsed;
      if (elapsed > 0)
        this->status.average_baudrate = bytesProcessed / elapsed;
      this->status.instant_baudrate   = this->status.average_baudrate;

      this->monitorCallback(this->monitorCallbackData,
                            this->source.c_str(), this->destination.c_str(),
                            NULL);
    }
  }

  bool ShouldCancel()
  {
    return gfal2_is_canceled(this->context);
  }

private:
  gfal2_context_t    context;
  gfalt_params_t     params;
  gfalt_monitor_func monitorCallback;

  gfalt_transfer_status_t monitorCallbackData;
  gfalt_hook_transfer_plugin_t status;
  time_t startTime;

  std::string source, destination;
};



#if GFAL2_VERSION_MAJOR >=2 && GFAL2_VERSION_MINOR >= 6
int gfal_xrootd_3rdcopy_check(plugin_handle plugin_data,
        gfal_context_t context, const char* src, const char* dst,
        gfal_url2_check check)
#else
int gfal_xrootd_3rdcopy_check(plugin_handle plugin_data,
         const char* src, const char* dst,
         gfal_url2_check check)
#endif
{
  if (check != GFAL_FILE_COPY) return 0;

  return (strncmp(src, "root://", 7) == 0 &&
          strncmp(dst, "root://", 7) == 0);
}



static void gfal_xrootd_3rd_init_url(gfal2_context_t context,
        XrdCl::URL& xurl, const char* url, const char* token)
{
  xurl.FromString(normalize_url(context, url));
  if (token) {
    XrdCl::URL::ParamsMap params;
    params.insert(std::make_pair("svcClass", token));
    xurl.SetParams(params);
  }
}


int gfal_xrootd_3rd_copy(plugin_handle plugin_data, gfal2_context_t context,
                         gfalt_params_t params,
                         const char* src, const char* dst,
                         GError** err)
{
  GError* internalError = NULL;

  XrdCl::URL source_url, dest_url;
  gfal_xrootd_3rd_init_url(context, source_url, src, gfalt_get_src_spacetoken(params, NULL));
  gfal_xrootd_3rd_init_url(context, dest_url, dst, gfalt_get_dst_spacetoken(params, NULL));

#if XrdMajorVNUM(XrdVNUMBER) == 4
  XrdCl::PropertyList job;

  job.Set("source", source_url.GetURL());
  job.Set("target", dest_url.GetURL());
  job.Set("force", gfalt_get_replace_existing_file(params, NULL));
  job.Set("posc", true);
  job.Set("thirdParty", "only");
  job.Set("tpcTimeout", gfalt_get_timeout(params, NULL));
#else
  XrdCl::JobDescriptor job;

  job.force              = gfalt_get_replace_existing_file(params, NULL);;
  job.posc               = true;
  job.thirdParty         = true;
  job.thirdPartyFallBack = false;
  job.checkSumPrint      = false;
#endif

  char checksumType[64] = {0};
  char checksumValue[512] = {0};

  bool performChecksum = gfalt_get_checksum_check(params, NULL);

  if (performChecksum) {
    gfalt_get_user_defined_checksum(params,
                                    checksumType, sizeof(checksumType),
                                    checksumValue, sizeof(checksumValue),
                                    NULL);
    if (!checksumType[0] || !checksumValue[0]) {
      char* defaultChecksumType = gfal2_get_opt_string(context,
                                                       XROOTD_CONFIG_GROUP, XROOTD_DEFAULT_CHECKSUM,
                                                       &internalError);
      if (internalError) {
        g_set_error(err, xrootd_domain,
                    internalError->code,
                    "[%s] %s", __func__, internalError->message);
        g_error_free(internalError);
        return -1;
      }

      strncpy(checksumType, defaultChecksumType, sizeof(checksumType));
      g_free(defaultChecksumType);
    }

#if XrdMajorVNUM(XrdVNUMBER) == 4
    job.Set("checkSumMode", "end2end");
    job.Set("checkSumType", predefinedChecksumTypeToLower(checksumType));
    job.Set("checkSumPreset", checksumValue);
#else
    job.checkSumType   = predefinedChecksumTypeToLower(checksumType);
    job.checkSumPreset = checksumValue;
#endif
  }

  XrdCl::CopyProcess process;

#if XrdMajorVNUM(XrdVNUMBER) == 4
  XrdCl::PropertyList results;
  process.AddJob(job, &results);
#else
  process.AddJob(&job);
#endif

  XrdCl::XRootDStatus status = process.Prepare();
  if (!status.IsOK()) {
    g_set_error(err, 0, status.errNo,
                "[%s] Error on XrdCl::CopyProcess::Prepare(): %s",
                __func__, status.ToStr().c_str());
    return -1;
  }


  CopyFeedback feedback(context, params);
  status = process.Run(&feedback);

  if (!status.IsOK()) {
    g_set_error(err, 0, status.errNo,
                "[%s] Error on XrdCl::CopyProcess::Run(): %s",
                __func__, status.ToStr().c_str());
    return -1;
  }

  return 0;
}
